
"""
This is a Python library of utilities SUNDIALS developer may find useful.
Right now it consists of the following modules:

- `logs`: this module has functions for parsing logs produced by `SUNLogger`.

"""
